from abc import ABC
from dataclasses import dataclass  # NOQA

from fractal.core.models import Model
from fractal.core.repositories import Repository
from pydantic.dataclasses import dataclass


@dataclass
class Product(Model):
    id: str
    name: str


class ProductRepository(Repository[Product], ABC):
    pass



from dataclasses import make_dataclass

from fractal.core.specifications.account_id_specification import AccountIdSpecification


def test_account_id_specification():
    spec = AccountIdSpecification("abc")
    DC = make_dataclass("DC", [("account_id", int)])
    assert spec.is_satisfied_by(DC(**dict(account_id="abc")))



from fractal.contrib.gcp.firestore.repositories import FirestoreRepositoryMixin
from fractal.core.event_sourcing.event_store import EventStoreRepository
from fractal.core.event_sourcing.message import Message


class FirestoreEventStoreRepository(
    EventStoreRepository, FirestoreRepositoryMixin[Message]
):
    pass



from dataclasses import dataclass
from typing import Generic

from fractal.core.command_bus.command import Command
from fractal.core.repositories import Entity
from fractal.core.specifications.generic.specification import Specification


@dataclass
class AddEntityCommand(Generic[Entity], Command):
    entity: Entity


@dataclass
class UpdateEntityCommand(Generic[Entity], Command):
    id: str
    entity: Entity


@dataclass
class DeleteEntityCommand(Generic[Entity], Command):
    specification: Specification



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Collection, Iterator


def parse_item(field_op: str, value: Any) -> Specification:
    if "__" not in field_op:
        from fractal.core.specifications.generic.operators import EqualsSpecification

        return EqualsSpecification(field_op, value)
    else:
        field, op = field_op.split("__")
        specification = None
        if op == "equals":
            from fractal.core.specifications.generic.operators import (
                EqualsSpecification as specification,
            )
        elif op == "in":
            from fractal.core.specifications.generic.operators import (
                InSpecification as specification,
            )
        elif op == "contains":
            from fractal.core.specifications.generic.operators import (
                ContainsSpecification as specification,
            )
        elif op == "lt":
            from fractal.core.specifications.generic.operators import (
                LessThanSpecification as specification,
            )
        elif op == "lte":
            from fractal.core.specifications.generic.operators import (
                LessThanEqualSpecification as specification,
            )
        elif op == "gt":
            from fractal.core.specifications.generic.operators import (
                GreaterThanSpecification as specification,
            )
        elif op == "gte":
            from fractal.core.specifications.generic.operators import (
                GreaterThanEqualSpecification as specification,
            )
        if specification:
            return specification(field, value)


def parse(**kwargs) -> Iterator[Specification]:
    for field_op, value in kwargs.items():
        yield parse_item(field_op, value)


class Specification(ABC):
    @abstractmethod
    def is_satisfied_by(self, obj: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def to_collection(self) -> Collection:
        raise NotImplementedError

    def And(self, specification: "Specification") -> "Specification":
        from fractal.core.specifications.generic.collections import AndSpecification

        return AndSpecification([self, specification])

    def Or(self, specification: "Specification") -> "Specification":
        from fractal.core.specifications.generic.collections import OrSpecification

        return OrSpecification([self, specification])

    def __str__(self):
        return self.__class__.__name__

    @staticmethod
    def Not(specification: "Specification") -> "Specification":
        from fractal.core.specifications.generic.operators import NotSpecification

        return NotSpecification(specification)

    @staticmethod
    def parse(**kwargs):
        specs = list(parse(**kwargs))
        if len(specs) > 1:
            from fractal.core.specifications.generic.collections import AndSpecification

            return AndSpecification(specs)
        elif len(specs) == 1:
            return specs[0]
        return None



from collections import defaultdict
from typing import List

from fractal.core.command_bus.command import Command
from fractal.core.command_bus.command_handler import CommandHandler


class CommandBus:
    def __init__(self):
        self.handlers = defaultdict(list)

    def set_handlers(self, handlers: List[CommandHandler]):
        for handler in handlers:
            self.add_handler(handler)

    def add_handler(self, handler: CommandHandler):
        self.handlers[handler.command.__name__].append(handler)

    async def handle_async(self, command: Command):
        ret = {}
        for handler in self.handlers[command.__class__.__name__]:
            res = await handler.handle(command)
            if res is not None:
                ret[handler.__class__.__name__] = res
        return ret

    def handle(self, command: Command):
        ret = {}
        for handler in self.handlers[command.__class__.__name__]:
            res = handler.handle(command)
            if res is not None:
                ret[handler.__class__.__name__] = res
        return ret



import json
import os
import uuid
from typing import Iterator, Optional

from fractal.core.exceptions import ObjectNotFoundException
from fractal.core.repositories import Entity, FileRepository, Repository
from fractal.core.specifications.generic.operators import NotSpecification
from fractal.core.specifications.generic.specification import Specification
from fractal.core.specifications.id_specification import IdSpecification
from fractal.core.utils.json_encoder import EnhancedEncoder


class RootDirMixin(object):
    def __init__(self, root_dir: str, *args, **kwargs):
        super(RootDirMixin, self).__init__(*args, **kwargs)

        self.root_dir = root_dir


class FileRepositoryMixin(RootDirMixin, Repository[Entity]):
    @property
    def _filename(self) -> str:
        return os.path.join(self.root_dir, "db", f"{self.__class__.__name__}.txt")

    @property
    def _entities(self):
        if not os.path.exists(self._filename):
            return None
        with open(self._filename, "r") as fp:
            for line in fp.readlines():
                yield self.entity(**json.loads(line))

    def add(self, entity: Entity) -> Entity:
        with open(self._filename, "a") as fp:
            fp.write(json.dumps(entity.asdict(), cls=EnhancedEncoder) + "\n")
        return entity

    def update(self, entity: Entity, upsert=False) -> Entity:
        try:
            current = self.find_one(IdSpecification(entity.id))
        except ObjectNotFoundException:
            current = None
        if current or upsert:
            self.remove_one(IdSpecification(entity.id))
            return self.add(entity)

    def remove_one(self, specification: Specification):
        entities = list(self.find(NotSpecification(specification)))
        with open(self._filename, "w") as fp:
            fp.writelines(
                [json.dumps(e.asdict(), cls=EnhancedEncoder) + "\n" for e in entities]
            )

    def find_one(self, specification: Specification) -> Optional[Entity]:
        for entity in filter(
            lambda i: specification.is_satisfied_by(i), self._entities
        ):
            return entity
        if self.object_not_found_exception:
            raise self.object_not_found_exception
        raise ObjectNotFoundException(f"{self.entity.__name__} not found!")

    def find(self, specification: Optional[Specification] = None) -> Iterator[Entity]:
        if specification:
            entities = filter(
                lambda i: specification.is_satisfied_by(i), self._entities
            )
        else:
            entities = self._entities
        for entity in entities:
            yield entity

    def is_healthy(self) -> bool:
        return True


class FileFileRepositoryMixin(RootDirMixin, FileRepository):
    def upload_file(self, data: bytes, content_type: str, reference: str = "") -> str:
        if not reference:
            reference = str(uuid.uuid4())
        with open(os.path.join(self.root_dir, "media", reference), "wb") as fp:
            fp.write(data)
        return reference

    def get_file(self, reference: str) -> bytes:
        with open(os.path.join(self.root_dir, "media", reference), "rb") as fp:
            return fp.read()

    def delete_file(self, reference: str) -> bool:
        try:
            os.remove(os.path.join(self.root_dir, "media", reference))
            return True
        except Exception:
            return False



import importlib
import logging
import os
import pathlib
from types import FunctionType
from typing import List, Tuple


class ApplicationContext(object):
    instance = None
    settings = None
    registered_repositories = []
    registered_command_handlers = []
    registered_ingress_services = []
    registered_egress_services = []
    registered_internal_services = []

    def __new__(cls, dotenv=True, *args, **kwargs):
        if not isinstance(cls.instance, cls):
            cls.instance = object.__new__(cls, *args, **kwargs)
            if dotenv:
                from dotenv import load_dotenv

                load_dotenv()
            cls.instance.load()
        return cls.instance

    @classmethod
    def register_repository(cls, name):
        setattr(cls, name, None)

        def inner(select_repository):
            cls.registered_repositories.append((name, select_repository))

        return inner

    @classmethod
    def register_command_handler(cls, command_handler):
        cls.registered_command_handlers.append(command_handler)

    @classmethod
    def register_ingress_service(cls, name):
        def inner(service):
            cls.registered_ingress_services.append((name, service))

        return inner

    @classmethod
    def register_egress_service(cls, name):
        def inner(service):
            cls.registered_egress_services.append((name, service))

        return inner

    @classmethod
    def register_internal_service(cls, name):
        def inner(service):
            cls.registered_internal_services.append((name, service))

        return inner

    def load(self):
        from fractal.core.utils.loggers import init_logging

        init_logging(os.getenv("LOG_LEVEL", "INFO"))
        self.logger = logging.getLogger("app")
        self.repositories = set()
        self.repository_names = set()
        self.service_names = set()

        root_dir = pathlib.Path(self.settings.ROOT_DIR)
        for file_name in pathlib.Path(os.path.dirname(self.settings.BASE_DIR)).glob(
            "service/**/*.py"
        ):
            parts = file_name.parts[len(root_dir.parts) :]
            if parts[-1].startswith("_"):
                parts = parts[:-1]
            importlib.import_module(".".join(parts).replace(".py", ""))

        self.load_internal_services()
        self.load_repositories()
        self.load_egress_services()

        from fractal.core.event_sourcing.event_publisher import EventPublisher

        self.event_publisher = EventPublisher(self.load_event_projectors())
        self.load_command_bus()
        self.load_ingress_services()

        for repository in self.repositories:
            assert repository.is_healthy()
        for service in self.services:
            assert service.is_healthy()

    def reload(self):
        self.load()

    def adapters(self):
        for repository in self.repositories:
            yield repository
        for service in self.services:
            yield service

    def load_internal_services(self):
        """Load services for internal use of the domain."""
        for name, service in self.registered_internal_services:
            self.install_service(
                service(self.settings)
                if isinstance(service, FunctionType)
                else service,
                name=name,
            )

        if hasattr(self.settings, "SECRET_KEY") and self.settings.SECRET_KEY:
            from fractal.contrib.tokens.services import AutomaticJwtTokenService

            self.install_service(AutomaticJwtTokenService, name="token_service")
        else:
            from fractal.contrib.tokens.services import StaticTokenService

            self.install_service(StaticTokenService, name="token_service")

    def load_repositories(self):
        """Load repositories for data access"""
        for name, repository in self.registered_repositories:
            setattr(
                self,
                name,
                self.install_repository(
                    repository(self.settings)
                    if isinstance(repository, FunctionType)
                    else repository,
                    name=name,
                ),
            )

        if (
            hasattr(self.settings, "EVENT_STORE_BACKEND")
            and self.settings.EVENT_STORE_BACKEND == "firestore"
        ):
            from fractal.contrib.gcp.firestore.event_store import (
                FirestoreEventStoreRepository,
            )
            from fractal.core.event_sourcing.event_store import EventStoreRepository

            self.event_store_repository: EventStoreRepository = (
                FirestoreEventStoreRepository(self.settings)
            )

            from fractal.contrib.fastapi.utils.json_encoder import (
                BaseModelEnhancedEncoder,
            )
            from fractal.core.event_sourcing.event import BasicSendingEvent
            from fractal.core.event_sourcing.event_store import (
                EventStore,
                JsonEventStore,
            )
            from fractal.core.utils.subclasses import all_subclasses

            self.event_store: EventStore = JsonEventStore(
                event_store_repository=self.event_store_repository,
                events=all_subclasses(BasicSendingEvent),
                json_encoder=BaseModelEnhancedEncoder,
            )
        else:
            from fractal.core.event_sourcing.event_store import (
                EventStoreRepository,
                InMemoryEventStoreRepository,
            )

            self.event_store_repository: EventStoreRepository = (
                InMemoryEventStoreRepository()
            )

            from fractal.core.event_sourcing.event_store import (
                EventStore,
                ObjectEventStore,
            )

            self.event_store: EventStore = ObjectEventStore(
                event_store_repository=self.event_store_repository,
            )

    def load_egress_services(self):
        """Load services to external interfaces that are initiated by this service (outbound)"""
        for name, service in self.registered_egress_services:
            self.install_service(
                service(self.settings)
                if isinstance(service, FunctionType)
                else service,
                name=name,
            )

    def load_event_projectors(self):
        from fractal.core.event_sourcing.event import EventCommandMapper
        from fractal.core.event_sourcing.projectors.command_bus_projector import (
            CommandBusProjector,
        )
        from fractal.core.utils.subclasses import all_subclasses

        projectors = []

        self.command_bus_projector = CommandBusProjector(
            lambda: self.command_bus,
            all_subclasses(EventCommandMapper),
        )
        projectors.append(self.command_bus_projector)

        if getattr(self.settings, "EVENT_STORE_PROJECTOR", None):
            from fractal.core.event_sourcing.projectors.event_store_projector import (
                EventStoreProjector,
            )

            projectors.append(EventStoreProjector(self.event_store))

        if getattr(self.settings, "PRINT_PROJECTOR", None):
            from fractal.core.event_sourcing.projectors.print_projector import (
                PrintEventProjector,
            )

            projectors.append(PrintEventProjector())

        if gcp_project_id := getattr(self.settings, "GCP_PROJECT_ID", None):
            from fractal.contrib.gcp.pubsub.projectors import PubSubEventBusProjector

            projectors.append(
                PubSubEventBusProjector(
                    project_id=gcp_project_id,
                    topic=getattr(self.settings, "GCP_PUBSUB_TOPIC", ""),
                ),
            )

        return projectors

    def load_command_bus(self):
        from fractal.core.command_bus.command_bus import CommandBus

        self.command_bus = CommandBus()

        for handler in self.registered_command_handlers:
            handler.install(self)

    def load_ingress_services(self):
        """Load services to external interfaces that are initiated by the external services (inbound)"""
        for name, service in self.registered_ingress_services:
            self.install_service(
                service(self.settings)
                if isinstance(service, FunctionType)
                else service,
                name=name,
            )

    def install_repository(self, repository, *, name=""):
        if not name:
            from fractal.core.utils.string import camel_to_snake

            name = camel_to_snake(repository.__class__.__name__)
        self.repository_names.add(name)
        self.repositories.add(repository)
        return repository

    def install_service(self, service, *, name=""):
        if not name:
            from fractal.core.utils.string import camel_to_snake

            name = camel_to_snake(service.__name__)
        self.service_names.add(name)
        setattr(
            ApplicationContext, name, property(lambda self: next(service.install(self)))
        )

    @property
    def services(self):
        services = []
        for service_name in self.service_names:
            services.append(getattr(self, service_name))
        return services

    def get_parameters(self, parameters: List[str]) -> Tuple:
        for parameter in parameters:
            if not hasattr(self, parameter):
                from fractal import FractalException

                raise FractalException(
                    f"ApplicationContext does not provide '{parameter}'"
                )
        return tuple(getattr(self, p) for p in parameters)



import json
import uuid
from abc import ABC, abstractmethod
from calendar import timegm
from datetime import datetime
from typing import Dict
from urllib.request import urlopen

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from jose import ExpiredSignatureError, JWTError, jwt

from fractal.contrib.tokens.exceptions import (
    NotAllowedException,
    TokenExpiredException,
    TokenInvalidException,
)
from fractal.contrib.tokens.settings import (
    ACCESS_TOKEN_EXPIRATION_SECONDS,
    REFRESH_TOKEN_EXPIRATION_SECONDS,
)
from fractal.core.services import Service
from fractal.core.utils.application_context import ApplicationContext


class TokenService(Service):
    @abstractmethod
    def generate(
        self,
        payload: Dict,
        token_type: str = "access",
        seconds_valid: int = ACCESS_TOKEN_EXPIRATION_SECONDS,
    ) -> str:
        raise NotImplementedError

    def _prepare(
        self, payload: Dict, token_type: str, seconds_valid: int, issuer: str
    ) -> Dict:
        utcnow = timegm(datetime.utcnow().utctimetuple())
        if not seconds_valid:
            seconds_valid = (
                REFRESH_TOKEN_EXPIRATION_SECONDS
                if token_type == "refresh"
                else ACCESS_TOKEN_EXPIRATION_SECONDS
            )
        payload.update(
            {
                "iat": utcnow,
                "nbf": utcnow,
                "jti": str(uuid.uuid4()),
                "iss": issuer,
                "exp": utcnow + seconds_valid,
                "typ": token_type,
            }
        )
        return payload

    @abstractmethod
    def verify(self, token: str, *, typ: str):
        raise NotImplementedError

    @abstractmethod
    def decode(self, token: str):
        raise NotImplementedError

    @abstractmethod
    def get_unverified_claims(self, token: str):
        raise NotImplementedError


class DummyJsonTokenService(TokenService):
    def generate(
        self,
        payload: Dict,
        token_type: str = "access",
        seconds_valid: int = ACCESS_TOKEN_EXPIRATION_SECONDS,
    ) -> str:
        return json.dumps(payload)

    def verify(self, token: str, *, typ: str):
        try:
            return json.loads(token)
        except Exception:
            raise TokenInvalidException()

    def decode(self, token: str):
        return json.loads(token)

    def get_unverified_claims(self, token: str):
        return self.decode(token)


class DummyTokenService(TokenService):
    def generate(
        self,
        payload: Dict,
        token_type: str = "access",
        seconds_valid: int = ACCESS_TOKEN_EXPIRATION_SECONDS,
    ) -> str:
        return json.dumps(payload)

    def verify(self, token: str, *, typ: str):
        return dict(
            iss="dummy",
            sub=str(uuid.uuid4()),
            account=str(uuid.uuid4()),
            email="dummy@dummy.dummy",
            typ=typ,
        )

    def decode(self, token: str):
        return dict(
            iss="dummy",
            sub=str(uuid.uuid4()),
            account=str(uuid.uuid4()),
            email="dummy@dummy.dummy",
            typ="access",
        )

    def get_unverified_claims(self, token: str):
        return self.decode(token)


class StaticTokenService(TokenService):
    def generate(
        self,
        payload: Dict,
        token_type: str = "access",
        seconds_valid: int = ACCESS_TOKEN_EXPIRATION_SECONDS,
    ) -> str:
        return json.dumps(payload)

    def verify(self, token: str, *, typ: str):
        return dict(
            iss="dummy",
            sub="00000000-0000-0000-0000-000000000000",
            account="00000000-0000-0000-0000-000000000000",
            email="dummy@dummy.dummy",
            typ=typ,
            roles=["user", "admin"],
        )

    def decode(self, token: str):
        return dict(
            iss="dummy",
            sub="00000000-0000-0000-0000-000000000000",
            account="00000000-0000-0000-0000-000000000000",
            email="dummy@dummy.dummy",
            typ="access",
            roles=["user", "admin"],
        )

    def get_unverified_claims(self, token: str):
        return self.decode(token)


class JwtTokenService(TokenService, ABC):
    @abstractmethod
    def decode(self, token: str):
        raise NotImplementedError

    def verify(self, token: str, *, typ: str):
        try:
            payload = self.decode(token)
        except ExpiredSignatureError:
            raise TokenExpiredException()
        except JWTError:
            raise TokenInvalidException()
        if payload["typ"] != typ:
            raise TokenInvalidException()
        return payload


class SymmetricJwtTokenService(JwtTokenService):
    def __init__(self, issuer: str, secret: str):
        self.issuer = issuer
        self.secret = secret
        self.algorithm = "HS256"

    @classmethod
    def install(cls, context: ApplicationContext):
        app_name, app_env, app_domain, secret_key = context.settings.get_parameters(
            ["APP_NAME", "APP_ENV", "APP_DOMAIN", "SECRET_KEY"]
        )
        yield cls(
            f"{app_name}@{app_env}.{app_domain}",
            secret_key,
        )

    def generate(
        self,
        payload: Dict,
        token_type: str = "access",
        seconds_valid: int = ACCESS_TOKEN_EXPIRATION_SECONDS,
    ) -> str:
        return jwt.encode(
            self._prepare(payload, token_type, seconds_valid, self.issuer),
            self.secret,
            algorithm=self.algorithm,
        )

    def decode(self, token: str):
        return jwt.decode(token, self.secret, algorithms=self.algorithm)

    def get_unverified_claims(self, token: str):
        return jwt.get_unverified_claims(token)


class AsymmetricJwtTokenService(JwtTokenService):
    def __init__(self, issuer: str, private_key: str, public_key: str):
        self.issuer = issuer
        self.private_key = private_key
        self.public_key = public_key
        self.algorithm = "RS256"

    @classmethod
    def install(cls, context: ApplicationContext):
        app_name, app_env, app_domain, private_key, public_key = context.get_parameters(
            ["app_name", "app_env", "app_domain", "private_key", "public_key"]
        )
        yield cls(
            f"{app_name}@{app_env}.{app_domain}",
            private_key,
            public_key,
        )

    def generate(
        self,
        payload: Dict,
        token_type: str = "access",
        seconds_valid: int = ACCESS_TOKEN_EXPIRATION_SECONDS,
    ) -> str:
        return jwt.encode(
            self._prepare(payload, token_type, seconds_valid, self.issuer),
            self.private_key,
            algorithm=self.algorithm,
        )

    def decode(self, token: str):
        return jwt.decode(token, self.public_key, algorithms=self.algorithm)

    def get_unverified_claims(self, token: str):
        return jwt.get_unverified_claims(token)


class ExtendedAsymmetricJwtTokenService(AsymmetricJwtTokenService):
    def __init__(
        self, issuer: str, private_key: str, public_key: RSAPublicKey, kid: str
    ):
        super(ExtendedAsymmetricJwtTokenService, self).__init__(issuer, private_key, "")
        self.public_key = public_key
        self.kid = kid

    def decode(self, token: str):
        return jwt.decode(
            token, self.public_key, algorithms=self.algorithm, issuer=self.issuer
        )

    def generate(
        self,
        payload: Dict,
        token_type: str = "access",
        seconds_valid: int = ACCESS_TOKEN_EXPIRATION_SECONDS,
    ) -> str:
        return jwt.encode(
            self._prepare(payload, token_type, seconds_valid, self.issuer),
            self.private_key,
            algorithm=self.algorithm,
            headers={"kid": self.kid},
        )


class AutomaticJwtTokenService(JwtTokenService):
    def __init__(self, issuer: str, secret: str):
        self.issuer = issuer
        self.symmetric_token_service = SymmetricJwtTokenService(
            issuer=issuer,
            secret=secret,
        )

    @classmethod
    def install(cls, context: ApplicationContext):
        yield cls(
            context.settings.ACCOUNT_SERVICE_HOST,
            context.settings.SECRET_KEY,
        )

    def generate(
        self,
        payload: Dict,
        token_type: str = "access",
        seconds_valid: int = ACCESS_TOKEN_EXPIRATION_SECONDS,
    ) -> str:
        return self.symmetric_token_service.generate(payload, token_type, seconds_valid)

    def verify(self, token: str, *, typ: str):
        headers = jwt.get_unverified_headers(token)
        claims = jwt.get_unverified_claims(token)
        if headers["alg"] == "HS256":
            return self.symmetric_token_service.verify(token, typ=typ)
        if headers["alg"] == "RS256":
            jsonurl = urlopen(f"{claims['iss']}/public/keys")
            jwks = json.loads(jsonurl.read())
            for key in jwks:
                if key["id"] == headers["kid"]:
                    public_key = serialization.load_pem_public_key(
                        key["public_key"].encode("utf-8"), backend=default_backend()
                    )
                    asymmetric_token_service = ExtendedAsymmetricJwtTokenService(
                        issuer=self.issuer,
                        private_key="",
                        public_key=public_key,
                        kid=key["id"],
                    )
                    return asymmetric_token_service.verify(token, typ=typ)
        raise NotAllowedException("No permission!")

    def decode(self, token: str):
        ...

    def get_unverified_claims(self, token: str):
        return jwt.get_unverified_claims(token)



from typing import Dict, Optional


class ProcessScope:
    def __init__(self, data: Optional[Dict] = None):
        self.data = data or {}

    def __getattr__(self, item):
        return self.data.get(item)

    def __getitem__(self, item):
        return self.data.get(item)

    def __setitem__(self, key, value):
        self.data[key] = value

    def __repr__(self):
        return self.data.__repr__()

    def update(self, scope: "ProcessScope") -> "ProcessScope":
        self.data.update(scope.data)
        return self

    def get(self, key, default=None):
        return self.data.get(key, default)



from abc import ABC, abstractmethod
from typing import Generic, Iterator

from fractal.core.repositories import Entity
from fractal.core.specifications.generic.specification import Specification


class SortRepositoryMixin(Generic[Entity], ABC):
    @abstractmethod
    def find_sort(
        self, specification: Specification = None, *, order_by: str = "", limit: int = 0
    ) -> Iterator[Entity]:
        ...

    def find_one_sort(
        self, specification: Specification = None, *, order_by: str = ""
    ) -> Entity:
        for release in self.find_sort(
            specification=specification, order_by=order_by, limit=1
        ):
            return release


class InMemorySortRepositoryMixin(SortRepositoryMixin[Entity]):
    def find_sort(
        self, specification: Specification = None, *, order_by: str = "", limit: int = 0
    ) -> Iterator[Entity]:
        entities = self.find(specification)
        if order_by:
            if reverse := order_by.startswith("-"):
                order_by = order_by[1:]
            entities = sorted(
                entities, key=lambda i: getattr(i, order_by), reverse=reverse
            )
        for i, entity in enumerate(entities):
            if limit and i == limit:
                break
            yield entity



def test_is_healthy(service):
    assert service.is_healthy()


def test_install(empty_application_context):
    from fractal.core.services import Service

    assert next(Service.install(empty_application_context)).__class__ == Service



import pytest


@pytest.fixture
def fastapi_client(fastapi_app_with_default_routes):
    from starlette.testclient import TestClient

    return TestClient(fastapi_app_with_default_routes)


@pytest.fixture
def a_base_model_object():
    from pydantic import BaseModel

    class ABaseModelObject(BaseModel):
        id: str
        name: str = "default_name"

    return ABaseModelObject(id="1")


@pytest.fixture
def fastapi_app(settings):
    from fractal.contrib.fastapi.install import install_fastapi

    return install_fastapi(settings)


@pytest.fixture
def fastapi_app_with_default_routes(fastapi_app, token_service_application_context, settings):
    from fractal.contrib.fastapi.routers.default import inject_default_routes

    fastapi_app.include_router(
        inject_default_routes(token_service_application_context, settings), tags=["default"]
    )
    return fastapi_app


@pytest.fixture
def failing_service_class():
    from fractal.core.services import Service

    class FailingService(Service):
        def is_healthy(self) -> bool:
            raise NotImplementedError

    return FailingService


@pytest.fixture
def failing_service_fastapi_client(
    fastapi_app, token_service_application_context, failing_service_class, settings
):
    from starlette.testclient import TestClient

    from fractal.contrib.fastapi.routers.default import inject_default_routes

    token_service_application_context.install_service(failing_service_class)

    fastapi_app.include_router(inject_default_routes(token_service_application_context, settings), tags=["default"])

    return TestClient(fastapi_app)


@pytest.fixture
def failing_route_fastapi_client(fastapi_app):
    from fastapi import APIRouter
    from starlette.testclient import TestClient

    from fractal.contrib.fastapi.routers import Routes
    from fractal.core.exceptions import DomainException

    router = APIRouter()

    @router.get(Routes.ROOT)
    def read_root():
        raise DomainException(
            message="failing_route",
            code="FAILING_ROUTE",
            status_code=500,
        )

    fastapi_app.include_router(router, tags=["default"])

    return TestClient(fastapi_app)



from abc import ABC, abstractmethod
from typing import Dict, List, Optional, Type
from uuid import UUID

from fastapi import APIRouter, Depends, status
from pydantic import BaseModel

from fractal import Fractal
from fractal.contrib.fastapi.routers import Routes
from fractal.contrib.fastapi.routers.domain.models import AdapterInfo, Info
from fractal.contrib.fastapi.routers.tokens import get_payload, get_payload_roles
from fractal.contrib.tokens.fractal import DummyTokenServiceFractal
from fractal.contrib.tokens.models import TokenPayload, TokenPayloadRoles
from fractal.core.models import Model
from fractal.core.services import Service
from fractal.core.utils.application_context import ApplicationContext
from fractal.core.utils.settings import Settings


def inject_default_routes(
    context: ApplicationContext, settings: Optional[Settings] = None
):
    router = APIRouter()

    @router.get(Routes.ROOT)
    def read_root():
        return {
            "FastAPI": settings.APP_NAME
            if settings and hasattr(settings, "APP_NAME")
            else "Fractal Service",
        }

    @router.get(Routes.INFO, responses={200: {"model": Info}})
    def info(
        payload: TokenPayload = Depends(
            get_payload(DummyTokenServiceFractal(context, settings))
        ),
    ):
        adapters = list(context.adapters())
        data = [
            AdapterInfo(
                adapter=adapter.__class__.__name__,
                status=False,
            )
            for adapter in adapters
        ]
        for i in range(len(data)):
            try:
                data[i].status_ok = adapters[i].is_healthy()
            except Exception as e:
                context.logger.error(e)
                data[i].status_ok = False
        return data

    return router


class DefaultRestRouterService(Service, ABC):
    domain_entity_class: Type[Model]
    entities_route: str
    entity_route: str
    entity_contract: Type[BaseModel]
    create_entity_contract: Type[BaseModel]

    @abstractmethod
    def add_entity(
        self,
        contract: BaseModel,
        **kwargs,
    ):
        raise NotImplementedError

    @abstractmethod
    def find_entities(
        self,
        q: str = "",
        **kwargs,
    ):
        raise NotImplementedError

    @abstractmethod
    def get_entity(
        self,
        entity_id: UUID,
        **kwargs,
    ):
        raise NotImplementedError

    @abstractmethod
    def update_entity(
        self,
        entity_id: UUID,
        contract: BaseModel,
        **kwargs,
    ):
        raise NotImplementedError

    @abstractmethod
    def delete_entity(
        self,
        entity_id: UUID,
        **kwargs,
    ) -> Dict:
        raise NotImplementedError


class Contract(BaseModel, ABC):
    @abstractmethod
    def to_entity(self, **kwargs):
        raise NotImplementedError


class BasicRestRouterService(DefaultRestRouterService):
    def add_entity(
        self,
        contract: Contract,
        **kwargs,
    ):
        try:
            UUID(contract.id)
        except (ValueError, TypeError):
            contract.id = None
        _entity = contract.to_entity(
            user_id=kwargs.get("sub"), account_id=kwargs.get("account")
        )
        return self.ingress_service.add(_entity, str(kwargs.get("sub")))

    def find_entities(
        self,
        q: str = "",
        **kwargs,
    ):
        return self.ingress_service.find(str(kwargs.get("account")), q)

    def get_entity(
        self,
        entity_id: UUID,
        **kwargs,
    ):
        return self.ingress_service.get(str(entity_id), str(kwargs.get("account")))

    def update_entity(
        self,
        entity_id: UUID,
        contract: Contract,
        **kwargs,
    ):
        if _entity := self.get_entity(entity_id, **kwargs):
            _entity = _entity.update(contract.dict())
        else:
            _entity = contract.to_entity(
                id=entity_id,
                user_id=kwargs.get("sub"),
                account_id=kwargs.get("account"),
            )
        return self.ingress_service.update(
            str(entity_id), _entity, str(kwargs.get("sub"))
        )

    def delete_entity(
        self,
        entity_id: UUID,
        **kwargs,
    ) -> Dict:
        self.ingress_service.delete(
            str(entity_id), str(kwargs.get("sub")), str(kwargs.get("account"))
        )
        return {}


def inject_default_rest_routes(
    fractal: Fractal,
    *,
    router_service_class: Type[DefaultRestRouterService],
    roles: Dict[str, List[str]] = None,
):
    if roles is None:
        roles = {}

    router = APIRouter()

    @router.post(
        router_service_class().entities_route,
        response_model=router_service_class().domain_entity_class,
        status_code=status.HTTP_201_CREATED,
        name=f"Add {router_service_class().domain_entity_class.__name__}",
    )
    def add_entity(
        entity: router_service_class().create_entity_contract,
        payload: TokenPayloadRoles = Depends(
            get_payload_roles(fractal, roles=roles.get("add", ["user"]))
        ),
    ):
        return router_service_class().add_entity(
            contract=entity,
            **payload.dict(),
        )

    @router.get(
        router_service_class().entities_route,
        response_model=List[router_service_class().domain_entity_class],
        status_code=status.HTTP_200_OK,
        name=f"Find {router_service_class().domain_entity_class.__name__} entities",
    )
    def find_entities(
        q: Optional[str] = "",
        payload: TokenPayloadRoles = Depends(
            get_payload_roles(fractal, roles=roles.get("get", ["user"]))
        ),
    ):
        return router_service_class().find_entities(
            q=q,
            **payload.dict(),
        )

    @router.get(
        router_service_class().entity_route,
        response_model=router_service_class().domain_entity_class,
        status_code=status.HTTP_200_OK,
        name=f"Get {router_service_class().domain_entity_class.__name__} entity",
    )
    def get_entity(
        entity_id: UUID,
        payload: TokenPayloadRoles = Depends(
            get_payload_roles(fractal, roles=roles.get("get", ["user"]))
        ),
    ):
        return router_service_class().get_entity(
            entity_id=entity_id,
            **payload.dict(),
        )

    @router.put(
        router_service_class().entity_route,
        response_model=router_service_class().domain_entity_class,
        status_code=status.HTTP_200_OK,
        name=f"Update {router_service_class().domain_entity_class.__name__}",
    )
    def update_entity(
        entity_id: UUID,
        entity: router_service_class().entity_contract,
        payload: TokenPayloadRoles = Depends(
            get_payload_roles(fractal, roles=roles.get("update", ["user"]))
        ),
    ):
        return router_service_class().update_entity(
            entity_id=entity_id,
            contract=entity,
            **payload.dict(),
        )

    @router.delete(
        router_service_class().entity_route,
        response_model=Dict,
        status_code=status.HTTP_200_OK,
        name=f"Delete {router_service_class().domain_entity_class.__name__}",
    )
    def delete_entity(
        entity_id: UUID,
        payload: TokenPayloadRoles = Depends(
            get_payload_roles(fractal, roles=roles.get("delete", ["user"]))
        ),
    ) -> Dict:
        return router_service_class().delete_entity(
            entity_id=entity_id,
            **payload.dict(),
        )

    return router



from fractal import Fractal

from app.context import ApplicationContext
from app.settings import Settings


class ApplicationFractal(Fractal):
    settings = Settings()
    context = ApplicationContext()



import json


def test_default_routes_root(fastapi_client, settings):
    from fractal.contrib.fastapi.routers import Routes

    response = fastapi_client.get(Routes.ROOT)
    assert response.status_code == 200
    assert json.loads(response.content) == {"FastAPI": settings.APP_NAME}


def test_default_routes_info(fastapi_client, token):
    from fractal.contrib.fastapi.routers import Routes

    response = fastapi_client.get(Routes.INFO, headers={"Authorization": f"Bearer {token}"})
    assert response.status_code == 200
    assert {(i["adapter"], i["status_ok"]) for i in json.loads(response.content)} == {
        ("InMemoryRepository", True),
        ("FakeService", True),
        ("DummyJsonTokenService", True),
    }


def test_default_routes_info_token_error(failing_service_fastapi_client):
    from fractal.contrib.fastapi.routers import Routes

    response = failing_service_fastapi_client.get(Routes.INFO)
    assert response.status_code == 401


def test_default_routes_info_error(failing_service_fastapi_client, token):
    from fractal.contrib.fastapi.routers import Routes

    response = failing_service_fastapi_client.get(Routes.INFO, headers={"Authorization": f"Bearer {token}"})
    assert response.status_code == 200
    assert {(i["adapter"], i["status_ok"]) for i in json.loads(response.content)} == {
        ("InMemoryRepository", True),
        ("FakeService", True),
        ("DummyJsonTokenService", True),
        ("FailingService", False),
    }


def test_not_existing_route(failing_route_fastapi_client):
    from fractal.contrib.fastapi.routers import Routes

    response = failing_route_fastapi_client.get(Routes.INFO)
    assert response.status_code == 404


def test_failing_route(failing_route_fastapi_client):
    from fractal.contrib.fastapi.routers import Routes

    response = failing_route_fastapi_client.get(Routes.ROOT)
    assert response.status_code == 500
    assert json.loads(response.content) == {
        "code": "FAILING_ROUTE",
        "message": "failing_route",
    }



import pytest


@pytest.fixture
def equals_specification():
    from fractal.core.specifications.generic.operators import EqualsSpecification

    return EqualsSpecification("id", 1)


@pytest.fixture
def other_equals_specification():
    from fractal.core.specifications.generic.operators import EqualsSpecification

    return EqualsSpecification("name", "test")


@pytest.fixture
def or_specification(equals_specification, other_equals_specification):
    from fractal.core.specifications.generic.collections import OrSpecification

    return OrSpecification([equals_specification, other_equals_specification])


@pytest.fixture
def and_specification(equals_specification, other_equals_specification):
    from fractal.core.specifications.generic.collections import AndSpecification

    return AndSpecification([equals_specification, other_equals_specification])


@pytest.fixture
def in_specification(equals_specification, other_equals_specification):
    from fractal.core.specifications.generic.operators import InSpecification

    return InSpecification("field", [1, 2, 3])


@pytest.fixture
def less_than_specification(equals_specification, other_equals_specification):
    from fractal.core.specifications.generic.operators import LessThanSpecification

    return LessThanSpecification("id", 1)


@pytest.fixture
def less_than_equal_specification(equals_specification, other_equals_specification):
    from fractal.core.specifications.generic.operators import LessThanEqualSpecification

    return LessThanEqualSpecification("id", 1)


@pytest.fixture
def greater_than_specification(equals_specification, other_equals_specification):
    from fractal.core.specifications.generic.operators import GreaterThanSpecification

    return GreaterThanSpecification("id", 1)


@pytest.fixture
def greater_than_equal_specification(equals_specification, other_equals_specification):
    from fractal.core.specifications.generic.operators import (
        GreaterThanEqualSpecification,
    )

    return GreaterThanEqualSpecification("id", 1)


@pytest.fixture
def regex_string_match_specification(equals_specification, other_equals_specification):
    from fractal.core.specifications.generic.operators import (
        RegexStringMatchSpecification,
    )

    return RegexStringMatchSpecification("id", "abc")


@pytest.fixture
def dict_specification(equals_specification, other_equals_specification):
    from fractal.core.specifications.generic.specification import Specification

    class DictSpecification(Specification):
        def __init__(self, collection):
            self.collection = collection

        def is_satisfied_by(self, obj) -> bool:
            raise NotImplementedError

        def to_collection(self) -> dict:
            return self.collection

    return DictSpecification({"id": 1, "test": 2})



import pytest


def test_commit(event_sourced_repository, event_stream):
    event_sourced_repository.commit(event_stream)

    assert len(
        event_sourced_repository.event_store.event_store_repository.entities
    ) == len(event_stream.events)
    assert (
        list(
            event_sourced_repository.event_store.event_store_repository.entities.values()
        )[0].object_id
        == event_stream.events[0].object_id
    )


def test_add(event_sourced_repository, aggregate_root_object_with_recorded_event):
    event_sourced_repository.add(aggregate_root_object_with_recorded_event)

    assert (
        len(event_sourced_repository.event_store.event_store_repository.entities) == 1
    )


def test_add_error(event_sourced_repository, regular_object):
    from fractal.core.exceptions import AggregateRootError

    with pytest.raises(AggregateRootError):
        event_sourced_repository.add(regular_object)


def test_update(event_sourced_repository, aggregate_root_object_with_recorded_event):
    event_sourced_repository.update(aggregate_root_object_with_recorded_event)

    assert (
        len(event_sourced_repository.event_store.event_store_repository.entities) == 1
    )


def test_update_error(event_sourced_repository, regular_object):
    from fractal.core.exceptions import AggregateRootError

    with pytest.raises(AggregateRootError):
        event_sourced_repository.update(regular_object)


def test_is_healthy(event_sourced_repository):
    assert event_sourced_repository.is_healthy()



import pytest

specifications = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), {"id": {"$eq": 1}}),
    (
        pytest.lazy_fixture("or_specification"),
        {"$or": [{"id": {"$eq": 1}}, {"name": {"$eq": "test"}}]},
    ),
    (
        pytest.lazy_fixture("and_specification"),
        {"$and": [{"id": {"$eq": 1}}, {"name": {"$eq": "test"}}]},
    ),
    (
        pytest.lazy_fixture("dict_specification"),
        {"$and": [{"id": {"$eq": 1}}, {"test": {"$eq": 2}}]},
    ),
]


@pytest.mark.parametrize("specification, expected", specifications)
def test_build(mongo_specification_builder, specification, expected):
    assert mongo_specification_builder.build(specification) == expected



from __future__ import annotations

import dataclasses
import logging
from abc import ABC, abstractmethod
from typing import Dict, Generator, Generic, List, Optional, TypeVar, get_type_hints

from sqlalchemy import MetaData, Table, create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.exc import ArgumentError, IntegrityError
from sqlalchemy.orm import Mapper, Session, sessionmaker

from fractal.contrib.sqlalchemy.specifications import SqlAlchemyOrmSpecificationBuilder
from fractal.core.exceptions import DomainException
from fractal.core.repositories import Entity, Repository
from fractal.core.specifications.generic.specification import Specification
from fractal.core.specifications.id_specification import IdSpecification

EntityDao = TypeVar("EntityDao")


class UnknownListItemTypeException(DomainException):
    code = "UNKNOWN_LIST_ITEM_TYPE_EXCEPTION"
    status_code = 500


class IntegrityErrorException(DomainException):
    code = "INTEGRITY_ERROR_EXCEPTION"
    status_code = 500


class SqlAlchemyDao(ABC):
    @staticmethod
    @abstractmethod
    def mapper(meta: MetaData, foreign_keys: Dict[str, Mapper]) -> Mapper:
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def table(meta: MetaData) -> Table:
        raise NotImplementedError

    @staticmethod
    def from_domain(obj: Entity) -> SqlAlchemyDao:
        raise NotImplementedError


class DaoMapper(ABC):
    instance = None

    def __new__(cls, *args, **kwargs):
        if not isinstance(cls.instance, cls):
            cls.instance = object.__new__(cls)
            cls.instance.done = False
        return cls.instance

    def start_mappers(self, engine: Engine):
        if not self.done:
            meta = MetaData()
            self.application_mappers(meta)
            meta.create_all(engine)
            self.done = True

    @abstractmethod
    def application_mappers(self, meta: MetaData):
        raise NotImplementedError


class AbstractUnitOfWork(ABC):
    def __enter__(self) -> AbstractUnitOfWork:
        return self

    def __exit__(self, *args):
        self.rollback()

    @abstractmethod
    def commit(self):
        raise NotImplementedError

    @abstractmethod
    def rollback(self):
        raise NotImplementedError


class SqlAlchemyUnitOfWork(AbstractUnitOfWork):
    def __init__(self):
        self.session_factory = None

    def __enter__(self) -> AbstractUnitOfWork:
        self.session = self.session_factory()  # type: Session
        return super().__enter__()

    def __exit__(self, *args):
        super().__exit__(*args)
        self.session.close()

    def commit(self):
        self.session.commit()

    def rollback(self):
        self.session.rollback()


class SqlAlchemyRepositoryMixin(
    Generic[Entity, EntityDao], Repository[Entity], SqlAlchemyUnitOfWork
):
    entity = Entity
    entity_dao = EntityDao
    application_mapper = DaoMapper

    def __init__(self, connection_str: str):
        super().__init__()

        self.connection_str = connection_str
        engine = create_engine(
            self.connection_str,
        )

        self.application_mapper().start_mappers(engine)

        self.session_factory = sessionmaker(
            bind=engine,
            expire_on_commit=False,
        )

    def add(self, entity: Entity) -> Entity:
        return self.__add(entity, self.entity_dao)

    def __add(self, entity: Entity, entity_dao_class: SqlAlchemyDao) -> Entity:
        entity_dao = entity_dao_class.from_domain(entity)
        with self:
            try:
                self.session.add(entity_dao)
                self.commit()
            except IntegrityError as e:
                raise IntegrityErrorException(e)
        return entity

    def update(self, entity: Entity, *, upsert=False) -> Entity:
        return self.__update(entity, self.entity_dao, upsert=upsert)

    def __update(
        self, entity: Entity, entity_dao_class: SqlAlchemyDao, *, upsert=False
    ) -> Entity:
        """Recursive function"""
        with self:
            try:
                existing_entity_dao = self._find_one_raw(
                    IdSpecification(entity.id), entity_dao_class=entity_dao_class
                )
            except ArgumentError:
                raise UnknownListItemTypeException(
                    f"DAO '{entity_dao_class}' has an unknown list collection DAO, please add the type for the list."
                )
            if existing_entity_dao:
                self.__update_existing_record(
                    entity, entity_dao_class, existing_entity_dao
                )
                return entity
            elif upsert:
                return self.__add(entity, entity_dao_class)

    def __update_existing_record(self, entity, entity_dao_class, existing_entity_dao):
        updating_entity_dao = entity_dao_class.from_domain(entity)
        regular_fields = []
        list_fields = []

        for k, v in entity_dao_class.__annotations__.items():
            if hasattr(v, "__origin__") and v.__origin__ is list:
                list_fields.append(k)
            else:
                regular_fields.append(k)

        self.__update_main_entity(
            existing_entity_dao, regular_fields, updating_entity_dao
        )
        self.__update_compound_entities(entity, entity_dao_class, list_fields)
        self.commit()

    def __update_main_entity(
        self, existing_entity_dao, regular_fields, updating_entity_dao
    ):
        for k, v in updating_entity_dao.__dict__.items():
            if hasattr(existing_entity_dao, k) and k in regular_fields:
                setattr(existing_entity_dao, k, v)

    def __update_compound_entities(self, entity, entity_dao_class, list_fields):
        for field in list_fields:
            item_dao_class = get_type_hints(entity_dao_class)[field].__args__[0]

            # check for new items
            for item in getattr(entity, field):
                self.__update(item, item_dao_class, upsert=True)

            # check for items to delete
            foreign_key = getattr(entity_dao_class, field).expression.right
            items = list(
                self.session.query(item_dao_class).filter(foreign_key == entity.id)
            )
            item_ids = [item.id for item in getattr(entity, field)]
            for item_dao in items:
                if item_dao.id not in item_ids:
                    self.session.delete(item_dao)

    def remove_one(self, specification: Specification):
        entity = self._find_one_raw(specification)
        if entity:
            self.session.delete(entity)
            self.commit()

    def find_one(self, specification: Specification) -> Optional[Entity]:
        entity = self._find_one_raw(specification)
        if entity:
            return self._dao_to_domain(entity)

    def find(
        self, specification: Optional[Specification] = None
    ) -> Generator[Entity, None, None]:
        entities = self._find_raw(specification)

        if specification:
            entities = filter(lambda i: specification.is_satisfied_by(i), entities)
        for entity in entities:
            yield self._dao_to_domain(entity)

    def _dao_to_domain(self, entity: Entity):
        return self.__dao_to_domain(entity, self.entity, self.entity_dao)

    def __dao_to_domain(
        self, entity: Entity, domain_model: Entity, entity_dao: SqlAlchemyDao
    ):
        """Recursive function"""
        list_fields = []
        for k, v in entity_dao.__annotations__.items():
            if hasattr(v, "__origin__") and v.__origin__ is list:
                list_fields.append(k)
        d = entity.__dict__
        for field in list_fields:
            if hasattr(entity, field):
                item_domain_model = get_type_hints(domain_model)[field].__args__[0]
                item_entity_dao = get_type_hints(entity_dao)[field].__args__[0]
                d[field] = [
                    self.__dao_to_domain(sub_entity, item_domain_model, item_entity_dao)
                    for sub_entity in getattr(entity, field)
                ]
        fields = set(f.name for f in dataclasses.fields(domain_model))
        return domain_model(**{k: v for k, v in d.items() if k in fields})

    def _find_one_raw(
        self,
        specification: Specification,
        *,
        entity_dao_class: Optional[SqlAlchemyDao] = None,
    ) -> Optional[Entity]:
        entities = self._find_raw(specification, entity_dao_class=entity_dao_class)

        for entity in filter(lambda i: specification.is_satisfied_by(i), entities):
            return entity

    def _find_raw(
        self,
        specification: Optional[Specification],
        *,
        entity_dao_class: Optional[SqlAlchemyDao] = None,
    ) -> List[Entity]:
        _filter = {}
        if specification:
            _filter = SqlAlchemyOrmSpecificationBuilder.build(specification)
        if isinstance(_filter, list):
            entities = []
            for f in _filter:
                entities.extend(
                    list(
                        self.session.query(
                            entity_dao_class or self.entity_dao
                        ).filter_by(**dict(f))
                    )
                )
            return entities
        else:
            return self.session.query(entity_dao_class or self.entity_dao).filter_by(
                **dict(_filter)
            )

    def is_healthy(self) -> bool:
        try:
            with self:
                self.session.execute("SELECT 1")
        except Exception as e:
            logging.exception(f"Database is unhealthy! {e}")
            return False
        return True



from fractal import Settings


class SettingsMixin(object):
    def __init__(self, settings: Settings, *args, **kwargs):
        super(SettingsMixin, self).__init__(*args, **kwargs)

        self.settings = settings



from dataclasses import make_dataclass

from fractal.core.specifications.generic.operators import (
    ContainsSpecification,
    EqualsSpecification,
    GreaterThanEqualSpecification,
    GreaterThanSpecification,
    InSpecification,
    LessThanEqualSpecification,
    LessThanSpecification,
    NotSpecification,
)


def test_not_specification():
    spec = NotSpecification(EqualsSpecification("id", 2))
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_in_specification():
    spec = InSpecification("id", [1, 2, 3])
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_equals_specification():
    spec = EqualsSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_less_than_specification():
    spec = LessThanSpecification("id", 2)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_less_than_equal_specification():
    spec = LessThanEqualSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_greater_than_specification():
    spec = GreaterThanSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=2)))


def test_greater_than_equal_specification():
    spec = GreaterThanEqualSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))


def test_contains_specification():
    spec = ContainsSpecification("name", "a")
    DC = make_dataclass("DC", [("name", str)])
    assert spec.is_satisfied_by(DC(**dict(name="fractal")))



import logging
from typing import Generator, Optional, Tuple

from pymongo import MongoClient
from pymongo.database import Database

from fractal.contrib.mongo.specifications import MongoSpecificationBuilder
from fractal.core.repositories import Entity, Repository
from fractal.core.specifications.generic.specification import Specification

logger = logging.getLogger("app")


def setup_mongo_connection(
    host, port, username, password, database
) -> Tuple[MongoClient, Database]:
    if host == "mongo-mock":
        import mongomock

        client = mongomock.MongoClient()
    else:
        if username:
            connection_string = f"mongodb+srv://{username}:{password}@{host}/{database}"
        else:
            connection_string = f"mongodb://{host}:{port}/{database}"
        connection_string += "?retryWrites=true&w=majority&connect=false"
        client = MongoClient(connection_string)
    db = client[database]
    return client, db


class MongoRepositoryMixin(Repository[Entity]):
    def __init__(self, host, port, username, password, database, collection):
        self.client, self.db = setup_mongo_connection(
            host, port, username, password, database
        )
        self.collection = getattr(self.db, collection)

    def add(self, entity: Entity) -> Entity:
        self.collection.insert_one(entity.asdict())
        return entity

    def update(self, entity: Entity, upsert=False) -> Entity:
        obj = self.collection.find_one(
            dict(
                id=entity.id,
            )
        )
        if not obj:
            if upsert:
                return self.add(entity)
            return
        obj.update(entity.asdict())
        self.collection.update_one(
            dict(
                id=entity.id,
                account_id=entity.account_id,
            ),
            {"$set": obj},
        )
        return entity

    def remove_one(self, specification: Specification):
        self.collection.delete_one(MongoSpecificationBuilder.build(specification))

    def find_one(self, specification: Specification) -> Optional[Entity]:
        for obj in self.collection.find(MongoSpecificationBuilder.build(specification)):
            return self._obj_to_domain(obj)

    def find(
        self, specification: Specification = None
    ) -> Generator[Entity, None, None]:
        for obj in self.collection.find(MongoSpecificationBuilder.build(specification)):
            yield self._obj_to_domain(obj)

    def is_healthy(self) -> bool:
        ok = self.client.server_info().get("ok", False)
        if not ok:
            return False
        return True

    @staticmethod
    def _obj_to_domain(obj) -> Entity:
        raise NotImplementedError



from abc import abstractmethod
from typing import Any, Collection, List

from fractal.core.specifications.generic.specification import Specification


class CollectionSpecification(Specification):
    def __init__(self, specifications: List[Specification]):
        self.specifications = specifications

    @abstractmethod
    def is_satisfied_by(self, obj: Any) -> bool:
        raise NotImplementedError

    def to_collection(self) -> Collection:
        return self.specifications

    def __str__(self):
        return f"{self.__class__.__name__}({','.join(map(lambda s: str(s), self.specifications))})"

    def __eq__(self, other):
        return type(self) == type(other) and self.specifications == other.specifications


class AndSpecification(CollectionSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return all([spec.is_satisfied_by(obj) for spec in self.specifications])

    def And(self, specification: Specification) -> Specification:
        if isinstance(specification, AndSpecification):
            return AndSpecification(self.specifications + specification.specifications)
        else:
            return AndSpecification(self.specifications + [specification])


class OrSpecification(CollectionSpecification):
    def is_satisfied_by(self, obj: Any) -> bool:
        return any([spec.is_satisfied_by(obj) for spec in self.specifications])

    def Or(self, specification: Specification) -> Specification:
        if isinstance(specification, OrSpecification):
            return OrSpecification(self.specifications + specification.specifications)
        else:
            return OrSpecification(self.specifications + [specification])



from abc import ABC, abstractmethod
from typing import Generic, Iterator, Optional, TypeVar

from fractal.core.exceptions import DomainException
from fractal.core.specifications.generic.specification import Specification

Entity = TypeVar("Entity")


class Repository(Generic[Entity], ABC):
    entity: Optional[Entity] = None
    object_not_found_exception: Optional[DomainException] = None

    @abstractmethod
    def add(self, entity: Entity) -> Entity:
        raise NotImplementedError

    @abstractmethod
    def update(self, entity: Entity, upsert=False) -> Entity:
        raise NotImplementedError

    @abstractmethod
    def remove_one(self, specification: Specification):
        raise NotImplementedError

    @abstractmethod
    def find_one(self, specification: Specification) -> Optional[Entity]:
        raise NotImplementedError

    @abstractmethod
    def find(self, specification: Specification = None) -> Iterator[Entity]:
        raise NotImplementedError

    @abstractmethod
    def is_healthy(self) -> bool:
        raise NotImplementedError


class FileRepository(Generic[Entity], ABC):
    @abstractmethod
    def upload_file(self, data: bytes, content_type: str, reference: str = "") -> str:
        ...

    @abstractmethod
    def get_file(self, reference: str) -> bytes:
        ...

    @abstractmethod
    def delete_file(self, reference: str) -> bool:
        ...



from fractal.core.repositories.inmemory_repository_mixin import InMemoryRepositoryMixin

from app.domain.products import Product, ProductRepository


class InMemoryProductRepository(ProductRepository, InMemoryRepositoryMixin[Product]):
    pass



import datetime
import json
from dataclasses import asdict

from fractal.core.event_sourcing.event import SendingEvent
from fractal.core.event_sourcing.event_projector import EventProjector
from fractal.core.event_sourcing.message import Message
from fractal.core.utils.json_encoder import EnhancedEncoder


class PrintEventProjector(EventProjector):
    def project(self, id: str, event: SendingEvent):
        message = Message(
            id=id,
            occurred_on=datetime.datetime.now(tz=datetime.timezone.utc),
            event=event.__class__.__name__,
            data=event,
            object_id=str(event.object_id),
            aggregate_root_id=str(event.aggregate_root_id),
        )
        print(json.dumps(asdict(message), cls=EnhancedEncoder))



from typing import Any

from fractal.core.specifications.generic.operators import EqualsSpecification


class AccountIdSpecification(EqualsSpecification):
    def __init__(self, account_id: Any):
        super(AccountIdSpecification, self).__init__("account_id", account_id)



from typing import Optional

from fractal import Fractal
from fractal.core.utils.application_context import ApplicationContext
from fractal.core.utils.settings import Settings


class DummyTokenServiceFractal(Fractal):
    def __init__(self, context: ApplicationContext, settings: Optional[Settings]):
        self.context = context
        self.token_service = getattr(context, "token_service")
        self.settings = settings
        super(DummyTokenServiceFractal, self).__init__()



from fractal.core.utils.application_context import ApplicationContext as BaseContext

from app.settings import Settings


class ApplicationContext(BaseContext):
    def load_repositories(self):
        from app.domain.products import ProductRepository

        if Settings().PRODUCT_REPOSITORY_BACKEND == "sql":
            '''example: some sql adapter code'''
        elif Settings().PRODUCT_REPOSITORY_BACKEND == "file":
            '''example: some file adapter code'''
        else:
            from app.adapters.products import InMemoryProductRepository

            self.product_repository: ProductRepository = self.install_repository(
                InMemoryProductRepository(),
            )



import re
from typing import Collection, Optional

from fractal.core.exceptions import DomainException
from fractal.core.specifications.generic.collections import (
    AndSpecification,
    OrSpecification,
)
from fractal.core.specifications.generic.operators import (
    EqualsSpecification,
    GreaterThanEqualSpecification,
    GreaterThanSpecification,
    InSpecification,
    LessThanEqualSpecification,
    LessThanSpecification,
    RegexStringMatchSpecification,
)
from fractal.core.specifications.generic.specification import Specification


class SpecificationNotMappedToMongo(DomainException):
    code = "SPECIFICATION_NOT_MAPPED_TO_MONGO"
    status_code = 500


class MongoSpecificationBuilder:
    @staticmethod
    def build(specification: Specification = None) -> Optional[Collection]:
        if specification is None:
            return None
        elif isinstance(specification, AndSpecification):
            return {
                "$and": [
                    MongoSpecificationBuilder.build(spec)
                    for spec in specification.to_collection()
                ]
            }
        elif isinstance(specification, OrSpecification):
            return {
                "$or": [
                    MongoSpecificationBuilder.build(spec)
                    for spec in specification.to_collection()
                ]
            }
        elif isinstance(specification, InSpecification):
            return {specification.field: {"$in": specification.value}}
        elif isinstance(specification, EqualsSpecification):
            return {specification.field: {"$eq": specification.value}}
        elif isinstance(specification, LessThanSpecification):
            return {specification.field: {"$lt": specification.value}}
        elif isinstance(specification, LessThanEqualSpecification):
            return {specification.field: {"$lte": specification.value}}
        elif isinstance(specification, GreaterThanSpecification):
            return {specification.field: {"$gt": specification.value}}
        elif isinstance(specification, GreaterThanEqualSpecification):
            return {specification.field: {"$gte": specification.value}}
        elif isinstance(specification, RegexStringMatchSpecification):
            return {
                specification.field: {"$regex": f".*{re.escape(specification.value)}.*"}
            }
        elif isinstance(specification.to_collection(), dict):
            return {
                "$and": [
                    {key: {"$eq": value}}
                    for key, value in dict(specification.to_collection()).items()
                ]
            }
        raise SpecificationNotMappedToMongo(
            f"Specification '{specification}' not mapped to Mongo query."
        )



from typing import Collection, Optional

from fractal.core.exceptions import DomainException
from fractal.core.specifications.generic.collections import (
    AndSpecification,
    OrSpecification,
)
from fractal.core.specifications.generic.operators import EqualsSpecification
from fractal.core.specifications.generic.specification import Specification


class SpecificationNotMappedToSqlAlchemyOrm(DomainException):
    code = "SPECIFICATION_NOT_MAPPED_TO_SLQALCHEMY_ORM"
    status_code = 500


class SqlAlchemyOrmSpecificationBuilder:
    @staticmethod
    def build(specification: Specification = None) -> Optional[Collection]:
        if specification is None:
            return None
        elif isinstance(specification, OrSpecification):
            return [
                SqlAlchemyOrmSpecificationBuilder.build(spec)
                for spec in specification.to_collection()
            ]
        elif isinstance(specification, AndSpecification):
            return {
                k: v
                for spec in specification.to_collection()
                if (i := SqlAlchemyOrmSpecificationBuilder.build(spec))
                for k, v in dict(i).items()
                if isinstance(i, dict)
            }
        elif isinstance(specification, EqualsSpecification):
            return {specification.field: specification.value}
        elif isinstance(specification.to_collection(), dict):
            return specification.to_collection()
        raise SpecificationNotMappedToSqlAlchemyOrm(
            f"Specification '{specification}' not mapped to SqlAlchemy Orm query."
        )



import logging

try:
    import sentry_sdk
except ImportError:
    sentry = False
else:
    sentry = True

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse

from fractal.contrib.fastapi.exceptions.error_message import ErrorMessage
from fractal.core.exceptions import DomainException
from fractal.core.utils.settings import Settings


def install_fastapi(settings: Settings):
    app = FastAPI(
        title=getattr(settings, "APP_NAME", "Fractal service"),
        version=getattr(settings, "APP_VERSION", "0.1.0"),
        root_path=getattr(settings, "OPENAPI_PREFIX_PATH", ""),
        openapi_url="/openapi.json"
        if str(getattr(settings, "APIDOCS_ENABLED", "true")).lower()
        in ["1", "yes", "true"]
        else None,
    )

    app.add_middleware(
        CORSMiddleware,
        allow_origins=getattr(settings, "ALLOW_ORIGINS", "").split(","),
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    if sentry:
        if sentry_dsn := getattr(settings, "SENTRY_DSN", ""):
            sentry_sdk.init(dsn=sentry_dsn)

    @app.exception_handler(DomainException)
    def unicorn_domain_exception_handler(request: Request, exc: DomainException):
        logger = logging.getLogger("app")
        if isinstance(exc.status_code, int) and 400 <= exc.status_code < 500:
            logger.warning(f"{exc.code} - {exc.message}")
        else:
            logger.error(f"{exc.code} - {exc.message}")

        if sentry:
            sentry_sdk.capture_exception(exc)

        return JSONResponse(
            status_code=exc.status_code,
            content=ErrorMessage(
                code=exc.code,
                message=exc.message,
            ).dict(),
            headers=exc.headers,
        )

    @app.exception_handler(Exception)
    def unicorn_exception_handler(request: Request, exc: Exception):
        logging.error(exc)

        if sentry:
            sentry_sdk.capture_exception(exc)

        return JSONResponse(
            status_code=500,
            content=ErrorMessage(
                code=exc.__class__.__name__,
                message=str(exc),
            ).dict(),
        )

    return app



from typing import Any

from fractal.core.specifications.generic.operators import EqualsSpecification


class IdSpecification(EqualsSpecification):
    def __init__(self, id: Any):
        super(IdSpecification, self).__init__("id", id)



from fractal.core.exceptions import DomainException


class TokenInvalidException(DomainException):
    code = "TOKEN_INVALID"
    status_code = 401

    def __init__(self, message=None):
        if not message:
            message = "The supplied token is invalid!"
        super(TokenInvalidException, self).__init__(message)


class TokenExpiredException(DomainException):
    code = "TOKEN_EXPIRED"
    status_code = 401

    def __init__(self, message=None):
        if not message:
            message = "The supplied token is expired!"
        super(TokenExpiredException, self).__init__(message)


class NotAllowedException(DomainException):
    code = "NOT_ALLOWED_ERROR"
    status_code = 401



import uuid
from abc import ABC, abstractmethod
from dataclasses import asdict
from datetime import datetime
from json import JSONEncoder
from typing import List, Optional, Type

from fractal.core.event_sourcing.event import BasicSendingEvent
from fractal.core.event_sourcing.event_stream import EventStream
from fractal.core.event_sourcing.message import Message
from fractal.core.exceptions import DomainException
from fractal.core.repositories import Repository
from fractal.core.repositories.inmemory_repository_mixin import InMemoryRepositoryMixin
from fractal.core.specifications.generic.specification import Specification


class EventNotMappedError(DomainException):
    code = "EVENT_NOT_MAPPED_ERROR"
    status_code = 501

    def __init__(self, event: str):
        super(EventNotMappedError, self).__init__(
            f"Event '{event}' is not mapped to be loaded from the EventStore.",
        )


class EventStore(ABC):
    @abstractmethod
    def commit(self, event_stream: EventStream, aggregate: str, version: int):
        raise NotImplementedError

    @abstractmethod
    def get_event_stream(self) -> EventStream:
        raise NotImplementedError

    @abstractmethod
    def is_healthy(self) -> bool:
        raise NotImplementedError


class EventStoreRepository(Repository[Message], ABC):
    entity = Message


class InMemoryEventStoreRepository(
    EventStoreRepository, InMemoryRepositoryMixin[Message]
):
    pass


class BasicEventStore(EventStore, ABC):
    def __init__(self, event_store_repository: EventStoreRepository):
        self.event_store_repository = event_store_repository

    def is_healthy(self) -> bool:
        return self.event_store_repository.is_healthy()


class ObjectEventStore(BasicEventStore):
    def commit(self, event_stream: EventStream, aggregate: str, version: int):
        for event in event_stream.events:
            self.event_store_repository.add(
                Message(
                    id=str(uuid.uuid4()),
                    occurred_on=datetime.utcnow(),
                    event=event.__class__.__name__,
                    data=event,
                    object_id=event.object_id,
                    aggregate_root_id=event.aggregate_root_id,
                )
            )

    def get_event_stream(
        self, specification: Optional[Specification] = None
    ) -> EventStream:
        return EventStream(
            events=list(
                map(lambda m: m.data, self.event_store_repository.find(specification))
            )
        )


class DictEventStore(BasicEventStore):
    def __init__(
        self,
        event_store_repository: EventStoreRepository,
        events: List[Type[BasicSendingEvent]],
    ):
        super(DictEventStore, self).__init__(event_store_repository)
        self.events = {e.__name__: e for e in events}

    def commit(self, event_stream: EventStream, aggregate: str, version: int):
        for event in event_stream.events:
            self.event_store_repository.add(
                Message(
                    id=str(uuid.uuid4()),
                    occurred_on=datetime.utcnow(),
                    event=event.__class__.__name__,
                    data=asdict(event),
                    object_id=event.object_id,
                    aggregate_root_id=event.aggregate_root_id,
                )
            )

    def get_event_stream(
        self, specification: Optional[Specification] = None
    ) -> EventStream:
        events = []
        for m in self.event_store_repository.find(specification):
            if event := self.events.get(m.event, None):
                events.append(event(**m.data))
            else:
                raise EventNotMappedError(m.event)
        return EventStream(events=events)


class JsonEventStore(BasicEventStore):
    def __init__(
        self,
        event_store_repository: EventStoreRepository,
        events: List[Type[BasicSendingEvent]],
        json_encoder: Optional[Type[JSONEncoder]] = None,
    ):
        super(JsonEventStore, self).__init__(event_store_repository)
        self.events = {e.__name__: e for e in events}
        self.json_encoder = json_encoder

    def commit(self, event_stream: EventStream, aggregate: str, version: int):
        import json

        for event in event_stream.events:
            self.event_store_repository.add(
                Message(
                    id=str(uuid.uuid4()),
                    occurred_on=datetime.utcnow(),
                    event=event.__class__.__name__,
                    data=json.dumps(asdict(event), cls=self.json_encoder),
                    object_id=event.object_id,
                    aggregate_root_id=event.aggregate_root_id,
                )
            )

    def get_event_stream(
        self, specification: Optional[Specification] = None
    ) -> EventStream:
        import json

        events = []
        for m in self.event_store_repository.find(specification):
            if event := self.events.get(m.event, None):
                events.append(event(**json.loads(m.data)))
            else:
                raise EventNotMappedError(m.event)
        return EventStream(events=events)


class PickleEventStore(BasicEventStore):
    def __init__(
        self,
        event_store_repository: EventStoreRepository,
        events: List[Type[BasicSendingEvent]],
    ):
        super(PickleEventStore, self).__init__(event_store_repository)
        self.events = {e.__name__: e for e in events}

    def commit(self, event_stream: EventStream, aggregate: str, version: int):
        import pickle

        for event in event_stream.events:
            self.event_store_repository.add(
                Message(
                    id=str(uuid.uuid4()),
                    occurred_on=datetime.utcnow(),
                    event=event.__class__.__name__,
                    data=pickle.dumps(event),
                    object_id=event.object_id,
                    aggregate_root_id=event.aggregate_root_id,
                )
            )

    def get_event_stream(
        self, specification: Optional[Specification] = None
    ) -> EventStream:
        import pickle

        events = []
        for m in self.event_store_repository.find(specification):
            events.append(pickle.loads(m.data))
        return EventStream(events=events)



import re
import unicodedata


def camel_to_snake(name):
    name = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", name).lower()


def snake_to_camel(name):
    return "".join(word.title() for word in name.split("_"))


def camel_case_to_spaces(value):
    """
    Split CamelCase and convert to lowercase. Strip surrounding whitespace.
    """
    re_camel_case = re.compile(r"(((?<=[a-z])[A-Z])|([A-Z](?![A-Z]|$)))")
    return re_camel_case.sub(r" \1", value).strip().lower()


def slugify(value, allow_unicode=False):
    """
    Convert to ASCII if 'allow_unicode' is False. Convert spaces to hyphens.
    Remove characters that aren't alphanumerics, underscores, or hyphens.
    Convert to lowercase. Also strip leading and trailing whitespace.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
    value = re.sub(r"[^\w\s-]", "", value.lower()).strip()
    return re.sub(r"[-\s]+", "-", value)


def slug_is_valid(value, allow_unicode=False):
    if allow_unicode:
        slug_re = re.compile(r"^[-\w]+\Z")
    else:
        slug_re = re.compile(r"^[-a-zA-Z0-9_]+\Z")
    invalid_input = not slug_re.search(str(value))
    return not invalid_input



from typing import List, Optional

from pydantic.main import BaseModel


class AdapterInfo(BaseModel):
    adapter: str
    status_ok: Optional[bool]


class Info(BaseModel):
    adapters: List[AdapterInfo]



import json
from dataclasses import asdict
from typing import Dict, List, Tuple, Type

from fractal.core.repositories import Entity
from fractal.core.repositories.inmemory_repository_mixin import InMemoryRepositoryMixin
from fractal.core.utils.json_encoder import EnhancedEncoder


class ExternalDataInMemoryRepositoryMixin(InMemoryRepositoryMixin[Entity]):
    def __init__(self, klass: Type[Entity]):
        super(ExternalDataInMemoryRepositoryMixin, self).__init__()
        self.klass = klass

    def load_data_dict(self, data: Dict):
        key = self.klass.__name__.lower()
        self.entities = {e["id"]: self.klass(**e) for e in data.get(key, [])}

    def dump_data_dict(self) -> Tuple[str, List[Dict]]:
        return self.klass.__name__.lower(), [asdict(e) for e in self.entities.values()]

    def load_data_json(self, data: Dict):
        key = self.klass.__name__.lower()
        self.entities = {
            e["id"]: self.klass(**e) for e in json.loads(data.get(key, []))
        }

    def dump_data_json(self) -> Tuple[str, str]:
        _, data = self.dump_data_dict()
        return self.klass.__name__.lower(), json.dumps(data, cls=EnhancedEncoder)



from dataclasses import make_dataclass

from fractal.core.specifications.object_of_account_specification import (
    ObjectOfAccountSpecification,
)


def test_object_of_account_specification():
    spec = ObjectOfAccountSpecification("abc", "def")
    DC = make_dataclass("DC", [("id", str), ("account_id", str)])
    assert spec.is_satisfied_by(DC(**dict(id="abc", account_id="def")))



from typing import List

from fractal.core.event_sourcing.event import BasicSendingEvent
from fractal.core.event_sourcing.event_projector import EventProjector
from fractal.core.event_sourcing.event_stream import EventStream


class EventPublisher:
    def __init__(self, projectors: List[EventProjector]):
        self.projectors = projectors

    def publish_event(self, event: BasicSendingEvent):
        self._publish(
            EventStream(
                events=[
                    event,
                ]
            )
        )

    def publish_events(self, events: List[BasicSendingEvent]):
        self._publish(
            EventStream(
                events=events,
            )
        )

    def _publish(self, event_stream: EventStream):
        for event in event_stream.events:
            for projector in self.projectors:
                projector.project(event_stream.id, event)



import uuid

from google.cloud import storage

from fractal import Settings
from fractal.contrib.gcp import SettingsMixin
from fractal.core.repositories import Entity, Repository


def get_cloudstorage_client(settings: Settings):
    if not hasattr(settings, "cloudstorage_client"):
        settings.cloudstorage_client = storage.Client()
    return settings.cloudstorage_client


class CloudStorageRepositoryMixin(SettingsMixin, Repository[Entity]):
    """
    https://github.com/googleapis/python-storage/tree/main/samples/snippets
    """

    entity = Entity

    def __init__(self, *args, **kwargs):
        super(CloudStorageRepositoryMixin, self).__init__(*args, **kwargs)

        client = get_cloudstorage_client(self.settings)
        if app_name := getattr(self.settings, "APP_NAME"):
            self.bucket = client.bucket(
                f"{app_name.lower()}-{self.entity.__name__.lower()}"
            )
        else:
            self.bucket = client.bucket(self.entity.__name__.lower())

    def upload_file(self, data: bytes, content_type: str, reference: str = "") -> str:
        # https://github.com/googleapis/python-storage/blob/main/samples/snippets/storage_upload_from_memory.py
        if not reference:
            reference = str(uuid.uuid4())
        blob = self.bucket.blob(reference)
        blob.upload_from_string(data, content_type=content_type)
        return reference

    def get_file(self, reference: str) -> bytes:
        # https://github.com/googleapis/python-storage/blob/main/samples/snippets/storage_download_into_memory.py
        blob = self.bucket.blob(reference)
        return blob.download_as_string()

    def delete_file(self, reference: str) -> bool:
        # https://github.com/googleapis/python-storage/blob/main/samples/snippets/storage_delete_file.py
        blob = self.bucket.blob(reference)
        blob.delete()
        return True



from dataclasses import dataclass

import pytest


@dataclass
class AnObject:
    id: str
    name: str = "default_name"


@pytest.fixture
def an_object():
    return AnObject("1")


@pytest.fixture
def another_object():
    return AnObject("2", "another_name")


@pytest.fixture
def inmemory_repository():
    from fractal.core.repositories.inmemory_repository_mixin import (
        InMemoryRepositoryMixin,
    )

    class InMemoryRepository(InMemoryRepositoryMixin[AnObject]):
        pass

    return InMemoryRepository()


@pytest.fixture
def inmemory_filter_repository():
    from fractal.core.repositories.inmemory_repository_mixin import (
        InMemoryRepositoryMixin,
    )
    from fractal.core.repositories.filter_repository_mixin import FilterRepositoryMixin

    class InMemoryFilterRepository(InMemoryRepositoryMixin[AnObject], FilterRepositoryMixin[AnObject]):
        pass

    return InMemoryFilterRepository()


@pytest.fixture
def external_data_inmemory_repository():
    from fractal.core.repositories.external_data_inmemory_repository_mixin import (
        ExternalDataInMemoryRepositoryMixin,
    )

    class ExternalDataInMemoryRepository(ExternalDataInMemoryRepositoryMixin[AnObject]):
        pass

    return ExternalDataInMemoryRepository(AnObject)



from dataclasses import make_dataclass

from fractal.core.specifications.id_specification import IdSpecification


def test_id_specification():
    spec = IdSpecification(1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))



from typing import Collection, Optional

from fractal.core.exceptions import DomainException
from fractal.core.specifications.generic.collections import AndSpecification
from fractal.core.specifications.generic.operators import (
    ContainsSpecification,
    EqualsSpecification,
    GreaterThanEqualSpecification,
    GreaterThanSpecification,
    InSpecification,
    LessThanEqualSpecification,
    LessThanSpecification,
)
from fractal.core.specifications.generic.specification import Specification


class SpecificationNotMappedToFirestore(DomainException):
    code = "SPECIFICATION_NOT_MAPPED_TO_FIRESTORE"
    status_code = 500


class FirestoreSpecificationBuilder:
    @staticmethod
    def build(specification: Specification = None) -> Optional[Collection]:
        if specification is None:
            return None
        elif isinstance(specification, AndSpecification):
            return [
                FirestoreSpecificationBuilder.build(spec)
                for spec in specification.to_collection()
            ]
        elif isinstance(specification, ContainsSpecification):
            return specification.field, "array-contains", specification.value
        elif isinstance(specification, InSpecification):
            if not specification.value:
                return None
            return specification.field, "in", specification.value
        elif isinstance(specification, EqualsSpecification):
            return specification.field, "==", specification.value
        elif isinstance(specification, LessThanSpecification):
            return specification.field, "<", specification.value
        elif isinstance(specification, LessThanEqualSpecification):
            return specification.field, "<=", specification.value
        elif isinstance(specification, GreaterThanSpecification):
            return specification.field, ">", specification.value
        elif isinstance(specification, GreaterThanEqualSpecification):
            return specification.field, ">=", specification.value
        elif isinstance(specification.to_collection(), dict):
            return [
                (key, "==", value)
                for key, value in dict(specification.to_collection()).items()
            ]
        raise SpecificationNotMappedToFirestore(
            f"Specification '{specification}' not mapped to Firestore query."
        )



from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar

Command = TypeVar("Command")


class CommandHandler(Generic[Command], ABC):
    command: Type[Command] = None

    @abstractmethod
    def handle(self, command: Command):
        """Handle command, might return a value."""



import pytest

specifications = [
    (None, None),
    (pytest.lazy_fixture("equals_specification"), {"id": 1}),
    (pytest.lazy_fixture("or_specification"), [{"id": 1}, {"name": "test"}]),
    (pytest.lazy_fixture("and_specification"), {"id": 1, "name": "test"}),
    (pytest.lazy_fixture("dict_specification"), {"id": 1, "test": 2}),
]


@pytest.mark.parametrize("specification, expected", specifications)
def test_build(sqlalchemy_specification_builder, specification, expected):
    assert sqlalchemy_specification_builder.build(specification) == expected


def test_build_error(sqlalchemy_specification_builder, greater_than_specification):
    from fractal.contrib.sqlalchemy.specifications import (
        SpecificationNotMappedToSqlAlchemyOrm,
    )

    with pytest.raises(SpecificationNotMappedToSqlAlchemyOrm):
        sqlalchemy_specification_builder.build(greater_than_specification)



import os

from fractal.core.utils.settings import Settings as BaseSettings


class Settings(BaseSettings):
    BASE_DIR = os.path.dirname(__file__)
    ROOT_DIR = os.path.dirname(os.path.dirname(BASE_DIR))
    APP_NAME = os.getenv("APP_NAME", "product_system")

    def load(self):
        self.PRODUCT_REPOSITORY_BACKEND = os.getenv("PRODUCT_REPOSITORY_BACKEND", "")



class DomainException(Exception):
    def __init__(
        self, message, code=None, status_code=None, payload=None, headers=None
    ):
        super(DomainException, self).__init__(message)
        self.message = message
        if status_code:
            self.status_code = status_code
        if code:
            self.code = code
        self.payload = payload
        self.headers = headers


class AggregateRootError(DomainException):
    code = "AGGREGATE_ROOT_ERROR"
    status_code = 405

    def __init__(self, message=None):
        if not message:
            message = "Actions can only be taken on aggregate root objects."
        super(AggregateRootError, self).__init__(message)


class ObjectNotFoundException(DomainException):
    code = "OBJECT_NOT_FOUND"
    status_code = 404

    def __init__(self, message=None):
        if not message:
            message = "Object not found!"
        super(ObjectNotFoundException, self).__init__(message)



from dataclasses import make_dataclass

from fractal.core.specifications.generic.operators import EqualsSpecification


def test_equals_specification():
    spec = EqualsSpecification("id", 1)
    DC = make_dataclass("DC", [("id", int)])
    assert spec.is_satisfied_by(DC(**dict(id=1)))



def test_load(fake_application_context_class, fake_service_class):
    assert not fake_application_context_class.instance
    assert not hasattr(
        fake_application_context_class.instance, fake_service_class.__class__.__name__
    )

    context = fake_application_context_class()
    assert context == fake_application_context_class.instance

    assert len(context.repositories) == 1
    assert len(context.services) == 2

    from fractal.core.utils.string import camel_to_snake

    assert hasattr(context, camel_to_snake(fake_service_class.__name__))


def test_reload(
    fake_application_context_class,
    fake_service_class,
    another_fake_service_class,
    settings,
):
    context = fake_application_context_class()

    assert type(context.fake_service) == fake_service_class

    settings.reload(
        {
            "FAKE_SERVICE": "another",
        }
    )
    context.reload()
    assert context == fake_application_context_class.instance

    assert len(context.repositories) == 1
    assert len(context.services) == 2

    from fractal.core.utils.string import camel_to_snake

    assert hasattr(context, camel_to_snake(fake_service_class.__name__))
    assert type(context.fake_service) == another_fake_service_class


def test_adapters(
    fake_application_context_class, inmemory_repository, fake_service_class, settings
):
    settings.reload(
        {
            "FAKE_SERVICE": "",
        }
    )
    context = fake_application_context_class()

    assert len(list(context.adapters())) == 3

    from fractal.contrib.tokens.services import StaticTokenService

    assert {type(a) for a in context.adapters()} == {inmemory_repository.__class__, fake_service_class, StaticTokenService}
